import os
import shutil
import hashlib

def get_hash_from_path(path, algorithm='md4'):
    """
    Returns the hash of the file `path`.
    """
    f = open(path)
    content = f.read()
    f.close()
    m = hashlib.new(algorithm)
    m.update(content)
    return m.hexdigest()

def copy_file(src, dst):
    """
    Copy `src` to `dst`.

    The parent directories for `dst` are created.

    To increase performance, this function will check if the file `dst`
    exists and compare the md4 hash of `src` and `dst`. The file will
    only be copied if the hashes differ.
    """
    try:
        os.makedirs(os.path.dirname(dst))
    except OSError:
        pass

    if os.path.isfile(dst):
        src_hash = get_hash_from_path(src)
        dst_hash = get_hash_from_path(dst)
        if src_hash == dst_hash:
            return
    try:
        shutil.copy(src, dst)
    except IOError:
        pass




import datetime
import re
from os.path import splitext


class Parser(object):
    output_ext = None

    def __init__(self, source):
        self.source = source
        self.headers = {}
        self.text = ''

    def _parse_headers(self):
        """
        Parses and removes the headers from the source.
        """
        META_RE = re.compile(
            r'^[ ]{0,3}(?P<key>[A-Za-z0-9_-]+):\s*(?P<value>.*)')
        lines = self.source.splitlines()
        for num, line in enumerate(lines):
            match = META_RE.match(line)
            if match:
                key = match.group('key').strip().lower()
                value = match.group('value').strip()
                if value:
                    # custom header transformation
                    header_method = getattr(self, '_parse_%s_header' % key,
                                            None)
                    if header_method:
                        value = header_method(value)
                    self.headers[key] = value
                    num_last_match = num
            else:
                break
        # remove header lines from input source
        try:
            del lines[:num_last_match+1]
        except UnboundLocalError:
            pass

        # check if a blank line followed the header lines and remove it
        try:
            if not lines[0]:
                del lines[0]
        except IndexError:
            pass
        self.text = '\n'.join(lines)

    def _parse_tags_header(self, value):
        """
        Parses the value from the 'tags' header into a list.
        """
        return [t.strip() for t in value.split(',')]

    def _parse_date_header(self, value):
        """
        Parses the date header into a python datetime.date object.
        The value must be in the 'Y-m-d' format.
        """
        try:
            return datetime.datetime.strptime(value, '%Y-%m-%d')
        except ValueError:
            return value

    def _parse_updated_header(self, value):
        return self._parse_date_header(value)

    def _parse_status_header(self, value):
        """
        Checks that the value of the 'status' header is 'live', 'hidden' or
        'draft'. If not 'live' is returned.
        """
        if value in ('live', 'hidden', 'draft'):
            return value
        return 'live'

    def _parse_text(self):
        """
        Returns the raw input text. Override this method to process
        text in another markup language such as Markdown.
        """
        return self.text

    def parse(self):
        self._parse_headers()
        self._parse_text()
        return (self.headers, self.text)


class RstParser(Parser):
    """ReStructuredText Parser"""
    output_ext = 'html'

    def pygments_directive(self, name, arguments, options, content, lineno,
                           content_offset, block_text, state, state_machine):
        """
        Parse sourcecode using Pygments
        
        From http://bitbucket.org/birkenfeld/pygments-main/src/tip/external/rst-directive-old.py
        """
        from pygments import highlight
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import get_lexer_by_name, TextLexer
        from docutils import nodes

        try:
            lexer = get_lexer_by_name(arguments[0])
        except ValueError:
            # no lexer found - use the text one instead of an exception
            lexer = TextLexer()
        # take an arbitrary option if more than one is given
        formatter = HtmlFormatter(noclasses=False)
        parsed = highlight(u'\n'.join(content), lexer, formatter)
        return [nodes.raw('', parsed, format='html')]
    pygments_directive.arguments = (1, 0, 1)
    pygments_directive.content = 1

    def _parse_text(self):
        try:
            from docutils.core import publish_parts
            from docutils.parsers.rst import directives
        except ImportError:
            raise Exception("The Python docutils library isn't installed.")
        else:
            # if pygments is installed, register the "sourcecode" directive
            try:
                import pygments
            except ImportError:
                pass
            else:
                directives.register_directive('sourcecode',
                                              self.pygments_directive)
            self.text = publish_parts(source=self.text,
                                      settings_overrides={
                                            "doctitle_xform": False,
                                      },
                                      writer_name='html4css1')['fragment']

class MarkdownParser(Parser):
    """Markdown Parser"""
    output_ext = 'html'
    
    def _parse_text(self):
        try:
            import markdown
        except ImportError:
            raise Exception("The Python markdown library isn't installed.")
        else:
            self.text = markdown.markdown(self.text,
                                          ['codehilite(css_class=highlight)'])


# a mapping of file extensions to the corresponding parser class
parser_map = (
    (('html', 'htm', 'xml'), Parser),
    (('rst',), RstParser),
    (('md', 'markdown'), MarkdownParser),
)

def get_parser_for_filename(filename):
    """
    Factory function returning a parser class based on the file extension.
    """
    ext = splitext(filename)[1][1:]
    try:
        return [pair[1] for pair in parser_map if ext in pair[0]][0]
    except IndexError:
        return




import os
import posixpath
from SimpleHTTPServer import SimpleHTTPRequestHandler
from BaseHTTPServer import HTTPServer
import urllib


class LanyonHTTPRequestHandler(SimpleHTTPRequestHandler):

    def translate_path(self, path):
        """
        The default behavior of SimpleHTTPRequestHandler.translate_path
        is to serve files from the current directory and all of the 
        directories below it. This class overrides the method and 
        looks for a `rootpath` attribute, which will be used instead
        of the current directory.
        """
        # abandon query parameters
        path = path.split('?',1)[0]
        path = path.split('#',1)[0]
        path = posixpath.normpath(urllib.unquote(path))
        words = path.split('/')
        words = filter(None, words)
        try:
            path = self.rootpath
        except AttributeError:
            path = os.getcwd()
        for word in words:
            drive, word = os.path.splitdrive(word)
            head, word = os.path.split(word)
            if word in (os.curdir, os.pardir): continue
            path = os.path.join(path, word)
        return path




from jinja2 import Environment, ChoiceLoader, FileSystemLoader, PackageLoader


class Jinja2Template(object):
    default_template = 'default.html'

    def __init__(self, settings):
        self.settings = settings
        self.env = Environment(loader=ChoiceLoader([
            FileSystemLoader(self.settings['template_dir']),
            PackageLoader('lanyon')]))
        self.env.filters['datetimeformat'] = self.datetimeformat

    def datetimeformat(self, value, format='%H:%M / %d-%m-%Y'):
        return value.strftime(format)

    def render_string(self, template_str, **kwargs):
        """Use `template_str` as a template"""
        template = self.env.from_string(template_str)
        return template.render(**kwargs)

    def render(self, template_name, **kwargs):
        """Use `template_name` as a template"""
        template = self.env.get_template(template_name)
        return template.render(**kwargs)




import BaseHTTPServer
from codecs import open
from collections import defaultdict, namedtuple
from datetime import datetime
import logging
from optparse import OptionParser
from os import walk, makedirs, error, getcwd
from os.path import splitext, join, dirname, split, abspath, getctime, isdir,\
                    basename, exists, relpath
from shutil import rmtree

from lanyon import __version__, parser
from lanyon.utils import copy_file
from lanyon.template import Jinja2Template
from lanyon.server import LanyonHTTPRequestHandler

class Site(object):

    def __init__(self, settings):
        self.settings = settings
        self.articles = []
        self.media = []
        self.build_time = datetime.today()

    def _get_default_headers(self, path):
        """
        Returns a dictionary with the default headers for an article.

        `title` - titleized version of the filename
        `date` - set to ctime. On unix this is the time of the most recent
                 metadata change; on windows the creation time
        `status` - set to 'live'
        `template` - set to 'default.html'
        """
        name = split(splitext(path)[0])[1]
        title = name.title()
        # todo: check if ctime is successfull
        date = datetime.fromtimestamp(getctime(path))
        return dict(title=title, date=date, status='live',
                    template='default.html')

    def _get_output_path(self, input_path, url, output_ext):
        """
        Returns the absolute output path for an article.

        `input_path` - absolute path of the input file
        `url` - parsed "url" header. This may be None if the article
                hasn't specified such a header.
        `output_ext` - file extension of the output file as specified
                       by the parser.
        """
        if url:
            if not basename(url):
                # url is a directory
                output_path = join(url, 'index.html')
            else:
                # url is a filename
                output_path = url
        else:
            output_file = splitext(input_path)[0] + '.' + output_ext
            output_path = relpath(output_file, self.settings['input_dir'])
        return join(self.settings['output_dir'], output_path)

    def _parse(self, input_data):
        Article = namedtuple('Article', 'headers, body, input_path, output_path, media')
        for input_dir in input_data:
            articles, media  = input_data[input_dir]

            # special case: media in the root input dir is not
            # associated with any articles
            if input_dir == self.settings['input_dir']:
                self.media = media
                media = []
            for path in articles:
                f = open(path, 'r', encoding='utf8')
                parser_inst = parser.get_parser_for_filename(path)(f.read())
                f.close()
                parsed = parser_inst.parse()

                output_ext = parser_inst.output_ext or \
                             splitext(path)[1][1:]
                output_path = self._get_output_path(path,
                                                    parsed[0].get('url'),
                                                    output_ext)

                # replace default headers with headers from the article
                default_headers = self._get_default_headers(path)
                headers = dict(default_headers, **parsed[0])
                headers['url'] = relpath(output_path, self.settings['output_dir'])

                # special case: remove 'index.html' from url
                head, tail = split(headers['url'])
                if tail == 'index.html':
                    headers['url'] = head + '/'

                if headers['status'] != 'draft' and\
                   headers['date'] <= datetime.today():
                    self.articles.append(Article(headers=headers,
                                                 body=parsed[1],
                                                 input_path=path,
                                                 output_path=output_path,
                                                 media=media))

    def _copy_media(self):
        # media that isn't associated with articles
        for medium in self.media:
            dst = join(self.settings['output_dir'],
                       relpath(medium, self.settings['input_dir']))
            copy_file(medium, dst)
        # media that is associated with articles
        for article in self.articles:
            for medium in article.media:
                dst = join(self.settings['output_dir'],
                           dirname(article.output_path),
                           relpath(medium, dirname(article.input_path)))
                copy_file(medium, dst)

    def _read_input(self):
        """
        Walks through the input dir and separates files into article
        and media files. Files and directories starting with a dot will
        be ignored.
        """
        data = defaultdict(lambda:([], []))

        for root, dirs, files in walk(self.settings['input_dir']):
            articles = []
            media = []

            # don't visit hidden directories
            for dir in dirs:
                if dir.startswith('.'):
                    dirs.remove(dir)

            # sort files into articles and media
            for file in files:
                if file.startswith('.'):
                    continue
                path = join(root, file)
                if parser.get_parser_for_filename(path):
                    articles.append(path)
                else:
                    media.append(path)

            if articles:
                data[root] = (articles, media)
            elif media:
                # dir has media file(s) but no articles. check if one of the parent 
                # dirs has an article and associate the media with it
                has_parent = False
                if root != self.settings['input_dir']:
                    parent_dir = dirname(root)
                    while parent_dir != self.settings['input_dir']:
                        if parent_dir in data:
                            data[parent_dir][1].extend(media)
                            has_parent = True
                        parent_dir = dirname(parent_dir)
                # if no parent dir could be found, or the file is in the
                # root dir associate the files with the root of the input dir
                if not has_parent:
                    data[self.settings['input_dir']][1].extend(media)
        return data

    def _is_public(self, article):
        """Return True if article is public"""
        if article.headers['status'] != 'hidden':
            return True
        return False

    def _sort_articles(self):
        """Sort articles by date"""
        self.articles.sort(key=lambda x: x.headers['date'], reverse=True)

    def _delete_output_dir(self):
        """Deletes the output directory"""
        if exists(self.settings['output_dir']):
            rmtree(self.settings['output_dir'])

    def _write(self):
        public_articles = filter(self._is_public, self.articles)
        template_cls = Jinja2Template(self.settings)
        for article in self.articles:
            try:
                makedirs(dirname(article.output_path))
            except OSError as e:
                logging.exception("error creating output directory")
                continue
            if article.headers['template'] == 'self':
                render_func = template_cls.render_string
                template = article.body
            else:
                render_func = template_cls.render
                template = article.headers['template']
            rendered = render_func(template,
                                   body=article.body,
                                   headers=article.headers,
                                   articles=public_articles,
                                   BUILD_TIME=self.build_time)
            fout = open(article.output_path, 'w', 'utf-8')
            fout.write(rendered)
            fout.close()

    def run(self):
        logging.info("project directory: %s", self.settings['project_dir']) 
        logging.info("reading input files")
        input_data = self._read_input()
        logging.info("parsing files")
        self._parse(input_data)
        self._sort_articles()
        logging.info("writing files")
        self._delete_output_dir()
        self._write()
        self._copy_media()
        logging.info("%s articles written", len(self.articles))

def main():
    parser = OptionParser(usage="%prog [project_dir]",
                                   version="%prog " + __version__)
    parser.add_option('-l', '--logging-level', help='Logging level')
    parser.add_option('-f', '--logging-file', help='Logging file name')
    parser.add_option('-p', '--port', help='Webserver port number',
                      default=8000, type='int')
    parser.add_option('-s', '--serve', help='Use local webserver',
                      action="store_true", dest="serve")
    (options, args) = parser.parse_args()

    LOGGING_LEVELS = {'critical': logging.CRITICAL,
                      'error': logging.ERROR,
                      'warning': logging.WARNING,
                      'info': logging.INFO,
                      'debug': logging.DEBUG}
    logging_level = LOGGING_LEVELS.get(options.logging_level, logging.INFO)
    logging.basicConfig(level=logging_level, filename=options.logging_file,
                        format='%(asctime)s %(levelname)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')
    try:
        project_dir = abspath(args[0])
    except IndexError:
        project_dir = abspath(getcwd())
    settings = {'project_dir': project_dir,
                'input_dir': join(project_dir, 'input'),
                'output_dir': join(project_dir, 'output'),
                'template_dir': join(project_dir, 'templates')}
    site = Site(settings)
    def runserver(server_class=BaseHTTPServer.HTTPServer,
                  handler_class=LanyonHTTPRequestHandler,
                  *args, **kwargs):
        site.run()
        handler_class.rootpath = settings['output_dir']
        server_address = ('', options.port)
        httpd = server_class(server_address, handler_class)
        logging.info("serving at port %s", server_address[1])
        httpd.serve_forever()
    if options.serve:
        import autoreload
        autoreload.main(runserver, (), {
            'paths': (settings['input_dir'], settings['template_dir'])
        })
    else:
        site.run()

if __name__ == '__main__':
    main()




# Autoreloading launcher.
# Borrowed from the Django project (http://www.djangoproject.com) which in turn
# borrowed it from Peter Hunt and the CherryPy project (http://www.cherrypy.org).
# Some taken from Ian Bicking's Paste (http://pythonpaste.org/).
#
# modified for lanyon (http://bitbucket.org/arthurk/lanyon/).
#
# Portions copyright (c) 2004, CherryPy Team (team@cherrypy.org)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of the CherryPy Team nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os, sys, time

try:
    import thread
except ImportError:
    import dummy_thread as thread

# This import does nothing, but it's necessary to avoid some race conditions
# in the threading module. See http://code.djangoproject.com/ticket/2330 .
try:
    import threading
except ImportError:
    pass

RUN_RELOADER = True

_win = (sys.platform == "win32")

def get_mtimes(directory):
    global _win
    mtimes = {}
    for root, dirs, files in os.walk(directory):
        for filename in files:
            filepath = os.path.join(root, filename)
            stat = os.stat(filepath)
            mtime = stat.st_mtime
            if _win:
                mtime -= stat.st_ctime
            mtimes[filepath] = mtime
    return mtimes

def files_changed(new_mtimes, initial_mtimes):
    """
    Compares `new_mtimes` with `initial_mtimes` and returns True if
    a key is not in both dicts or the value of a key is not the same.
    """
    for f in new_mtimes:
        if not f in initial_mtimes:
            # new file
            return True
        elif initial_mtimes[f] != new_mtimes[f]:
            # file modified
            return True
    for f in initial_mtimes:
        if not f in new_mtimes:
            # removed
            return True
    return False

def reloader_thread(args, kwargs):
    initial_mtimes = {}
    for mtimes in map(get_mtimes, kwargs['paths']):
        initial_mtimes.update(mtimes)
    while RUN_RELOADER:
        new_mtimes = {}
        for mtimes in map(get_mtimes, kwargs['paths']):
            new_mtimes.update(mtimes)
        if files_changed(new_mtimes, initial_mtimes):
            sys.exit(3) # force reload
        time.sleep(1)

def restart_with_reloader():
    while True:
        args = [sys.executable] + sys.argv
        if sys.platform == "win32":
            args = ['"%s"' % arg for arg in args]
        new_environ = os.environ.copy()
        new_environ["RUN_MAIN"] = 'true'
        exit_code = os.spawnve(os.P_WAIT, sys.executable, args, new_environ)
        if exit_code != 3:
            return exit_code

def python_reloader(main_func, args, kwargs):
    if os.environ.get("RUN_MAIN") == "true":
        thread.start_new_thread(main_func, args, kwargs)
        try:
            reloader_thread(args, kwargs)
        except KeyboardInterrupt:
            pass
    else:
        try:
            sys.exit(restart_with_reloader())
        except KeyboardInterrupt:
            pass

def jython_reloader(main_func, args, kwargs):
    from _systemrestart import SystemRestart
    thread.start_new_thread(main_func, args)
    while True:
        if files_changed(kwargs['directory']):
            raise SystemRestart
        time.sleep(1)


def main(main_func, args=None, kwargs=None):
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    if sys.platform.startswith('java'):
        reloader = jython_reloader
    else:
        reloader = python_reloader
    reloader(main_func, args, kwargs)




from imp import load_source
from os.path import join, dirname, abspath
from setuptools import setup, find_packages

lanyon_mod = load_source('lanyon', join(dirname(abspath(__file__)),
                                        'src', 'lanyon', '__init__.py'))

setup(
    name = 'lanyon',
    version = lanyon_mod.__version__,
    url = lanyon_mod.__url__,
    license = 'BSD',
    description = 'A static website generator that uses the file system structure to organize articles.',
    author = 'Arthur Koziel',

    packages = find_packages('src'),
    package_dir = {'': 'src'},
    package_data={'lanyon': ['templates/default.html'],},
    zip_safe = False,

    test_suite = 'nose.collector',

    # install the lanyon executable
    entry_points = {
        'console_scripts': [
            'lanyon = lanyon.app:main'
        ]
    },
    install_requires = [
        'Jinja2',
    ],
)

